import re
from collections import defaultdict
import ujson as json
import argparse
from tqdm import tqdm
from typing import TextIO
from datetime import datetime
import timeit
import time
import uuid
from xml.sax.saxutils import escape

def load_people(people_file: TextIO):
    cache = []
    max_idx = 0
    ppl = {}

    for line in people_file:
        # (John Hund, 73896)
        name, idx = re.match(r'^\((.*),(\d+)\)$', line).groups()

        # 0 is reserved for version
        # 1 is reserved for unknown node
        # 2 is reserved for unknown node alias
        idx = int(idx) + 3
        if max_idx < idx:
            max_idx = idx
        if idx == 2:
            cache.append(name)
        else:
            ppl[name] = idx
    for c in cache:
        max_idx += 1
        ppl[c] = max_idx

    idx2ppl = defaultdict(set)
    for name, idx in ppl.items():
        idx2ppl[idx].add(name)

    return ppl, idx2ppl, max_idx


def people2graphml(node_cnt, edge_cnt, idx2ppl, graphml_nodes, graphml_edges):
    graphml_nodes += ("<node id='1'>\n"
                      "  <data key='type'>person</data>\n"
                      "  <data key='pID'>1</data>\n"
                      "  <data key='labelV'>person</data>\n"
                      "</node>\n"
                      "<node id='2'>\n"
                      "  <data key='type'>alias</data>\n"
                      "  <data key='pID'>1</data>\n"
                      "  <data key='name'>UNKNOWN</data>\n"
                      "  <data key='labelV'>alias</data>\n"
                      "</node>\n")
    graphml_edges += ("<edge id='0' source='1' target='2'>\n"
                      "  <data key='labelE'>alias</data>\n"
                      "</edge>\n")

    for idx, names in idx2ppl.items():
        graphml_nodes += (f"<node id='{idx}'>\n"
                          f"  <data key='type'>person</data>\n"
                          f"  <data key='pID'>{idx}</data>\n"
                          f"  <data key='labelV'>person</data>\n"
                          f"</node>\n")
        for name in names:
            graphml_nodes += (f"<node id='{node_cnt}'>\n"
                              f"  <data key='type'>alias</data>\n"
                              f"  <data key='pID'>{idx}</data>\n"
                              f"  <data key='name'>{re.sub(r'[^a-zA-Z 0-9@.]', '', name)}</data>\n"
                              f"  <data key='labelV'>alias</data>\n"
                              f"</node>\n")

            graphml_edges += (f"<edge id='{edge_cnt}' source='{idx}' target='{node_cnt}'>\n"
                              f"  <data key='labelE'>alias</data>\n"
                              f"</edge>\n")
            node_cnt += 1
            edge_cnt += 1

    return node_cnt, edge_cnt, graphml_nodes, graphml_edges


def _get_ppl_id(name, email, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges):
    if name in ppl and bool(ppl[name]):
        idx = ppl[name]
    elif email in ppl and bool(ppl[email]):
        idx = ppl[email]
    else:
        graphml_nodes += (f"<node id='{node_cnt}'>\n"
                          f"  <data key='type'>person</data>\n"
                          f"  <data key='pID'>{node_cnt}</data>\n"
                          f"  <data key='labelV'>person</data>\n"
                          f"</node>\n")
        idx = node_cnt
        node_cnt += 1
        for alias in [name, email]:
            if bool(alias):
                ppl[name] = idx
                graphml_nodes += (f"<node id='{node_cnt}'>\n"
                                  f"  <data key='type'>alias</data>\n"
                                  f"  <data key='pID'>{idx}</data>\n"
                                  f"  <data key='name'>{re.sub(r'[^a-zA-Z 0-9@.]', '', alias)}</data>\n"
                                  f"  <data key='labelV'>alias</data>\n"
                                  f"</node>\n")
                graphml_edges += (f"<edge id='{edge_cnt}' source='{idx}' target='{node_cnt}'>\n"
                                  f"  <data key='labelE'>alias</data>\n"
                                  f"</edge>\n")
                node_cnt += 1
                edge_cnt += 1
    return idx, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges


def build_block_index(quagga_file: TextIO, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges):
    round_to = 5  # minutes
    block_cnt = 0
    index = {}
    blocks = []
    quagga_file.readline()
    for line in tqdm(quagga_file):
        quagga = json.loads(line)
        prev_block = None
        for block in quagga['quagga.parsed']['blocks']:
            s_id, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges = \
                _get_ppl_id(block['from']['name'], block['from']['email'],
                            ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges)

            try:
                dt = block['sent'] or ''
                dt = dt.replace('PDT', 'UTC-0700')
                dt = dt.replace('PST', 'UTC-0800')
                dt = dt.replace('LMT', 'UTC-0500')

                dt = datetime.strptime(dt, '%Y-%m-%d %H:%M:%S %Z%z')
                sent = datetime(dt.year, dt.month, dt.day, dt.hour,
                                min(round_to * round((float(dt.minute) + (float(dt.second) / 60)) / round_to), 59))
                sent_round = sent.strftime('%Y-%m-%d %H:%M:%S')
            except ValueError as e:
                if str(e) != "time data '' does not match format '%Y-%m-%d %H:%M:%S %Z%z'":
                    print(e)
                sent_round = str(uuid.uuid4())

            block_ = {
                'subject': re.sub(r'[^a-zA-Z 0-9@.]', '', block['subject'] or ''),
                'sent': block['sent'] or '',
                'sent_round': sent_round,
                'block_type': block['type'],
                'text': escape('\n'.join(block['text'])),
                'sender': s_id,
                'recipients': [],
                'recipients_cc': [],
                'block_id': block_cnt,
                'prev_block': prev_block
            }
            for recipient in (block['to'] or []):
                r_id, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges = \
                    _get_ppl_id(recipient['name'], recipient['email'],
                                ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges)
                block_['recipients'].append(r_id)
            for recipient in (block['cc'] or []):
                r_id, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges = \
                    _get_ppl_id(recipient['name'], recipient['email'],
                                ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges)
                block_['recipients_cc'].append(r_id)
            prev_block = block_cnt

            blocks.append(block_)
            if f'{s_id}||{sent_round}' not in index:
                index[f'{s_id}||{sent_round}'] = []
            index[f'{s_id}||{sent_round}'].append(block_cnt)
            block_cnt += 1

    return index, blocks, ppl, node_cnt, edge_cnt, graphml_nodes, graphml_edges


def quagga2graphml(blocks, block_index, node_cnt, edge_cnt, graphml_nodes, graphml_edges, include_text):
    block2node = {}
    orig_id_map = {}
    prev_blocks = []
    for block_key, block_ids in block_index.items():
        orig_id = block_ids[0]
        orig_cnt = node_cnt
        for block_i, block_id in enumerate(block_ids):
            block = blocks[block_id]
            is_duplicate = block_i > 0

            graphml_nodes += (f"<node id='{node_cnt}'>\n"
                              f"  <data key='type'>email</data>\n"
                              f"  <data key='subject'>{block['subject']}</data>\n"
                              f"  <data key='sent'>{block['sent']}</data>\n"
                              f"  <data key='block_type'>{block['block_type']}</data>\n"
                              f"  <data key='text'>{block['text'] if include_text else ''}</data>\n"
                              f"  <data key='labelV'>{'duplicate' if is_duplicate else 'email'}</data>\n"
                              f"  <data key='original'>{orig_cnt}</data>\n"
                              f"</node>\n")
            block2node[block_id] = node_cnt
            orig_id_map[block_id] = orig_id
            node_cnt += 1

            if block['prev_block']:
                prev_blocks.append((block['prev_block'], orig_cnt))

        graphml_edges += (f"<edge id='{edge_cnt}' source='{blocks[orig_id]['sender']}' target='{orig_cnt}'>\n"
                          f"  <data key='labelE'>sender</data>\n"
                          f"</edge>\n")
        edge_cnt += 1

        recipients = list(set([recipient for b_id in block_ids for recipient in blocks[b_id]['recipients']]))
        recipients_cc = list(set([recipient for b_id in block_ids for recipient in blocks[b_id]['recipients_cc']]))
        for recipient, r_type in zip(recipients + recipients_cc,
                                     (['to'] * len(recipients)) + (['to'] * len(recipients))):
            graphml_edges += (f"<edge id='{edge_cnt}' source='{orig_cnt}' target='{recipient}'>\n"
                              f"  <data key='labelE'>recipient</data>\n"
                              f"  <data key='rec_type'>{r_type}</data>\n"
                              f"</edge>\n")
            edge_cnt += 1

    for prev_block, target in prev_blocks:
        graphml_edges += (f"<edge id='{edge_cnt}'"
                          f"  source='{block2node[orig_id_map[prev_block]]}'"
                          f"  target='{target}'>\n"
                          f"  <data key='labelE'>conversation</data>\n"
                          f"</edge>\n")
        edge_cnt += 1

    return graphml_nodes, graphml_edges


def construct_graphml(graphml_nodes, graphml_edges):
    graphml = (""
               "<graphml xmlns='http://graphml.graphdrawing.org/xmlns'>\n"
               "  <key id='type'       for='node' attr.name='type'       attr.type='string'></key>\n"
               "  <key id='labelV'     for='node' attr.name='labelV'     attr.type='string'></key>\n"
               "  <key id='original'   for='node' attr.name='original'   attr.type='int'></key>\n"
               "  <key id='pID'        for='node' attr.name='pID'        attr.type='int'></key>\n"
               "  <key id='name'       for='node' attr.name='name'       attr.type='string'></key>\n"
               "  <key id='subject'    for='node' attr.name='subject'    attr.type='string'></key>\n"
               "  <key id='sent'       for='node' attr.name='sent'       attr.type='string'></key>\n"
               "  <key id='block_type' for='node' attr.name='block_type' attr.type='string'></key>\n"
               "  <key id='rec_type' for='edge' attr.name='rec_type' attr.type='string'></key>\n"
               "  <key id='labelE'   for='edge' attr.name='labelE'   attr.type='string'></key>\n"
               "  <graph id='routes' edgedefault='directed'>\n"
               "    <node id='0'>\n"
               "      <data key='labelV'>version</data>\n"
               "      <data key='type'>version</data>\n"
               "      <data key='code'>0.2</data>\n"
               "      <data key='desc'>Version: 0.1 Generated: 2018-12-04 23:24:40 UTC\n"
               "         Graph created by Tim Repke and Ben Hurdelhey\n"
               "         Based on Enron corpus.</data>\n"
               "    </node>\n" +
               graphml_nodes +
               graphml_edges +
               "  </graph>\n"
               "</graphml>")
    return graphml


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--people', type=argparse.FileType('r'),
                        help='File of deduplicated people with one alias per line: (Alias, person_id)')
    parser.add_argument('--quagga-parsed', type=argparse.FileType('r'),
                        help='File with one email per line parsed by quagga')
    parser.add_argument('--graph', type=argparse.FileType('w'),
                        help='GraphML file to write to.')
    parser.add_argument('--include-text', action='store_true',
                        help='add this flag to include the email text in the graph')
    args = parser.parse_args()

    print('Constructing GraphML File')
    print('Started at', time.strftime("%Y-%m-%d %H:%M"))
    print(args)

    time0 = timeit.default_timer()

    print('Loading index of people aliases...')
    time1 = timeit.default_timer()
    people, idx2people, last_idx = load_people(args.people)
    print(f' - done after {timeit.default_timer() - time1:.1f}s (total: {timeit.default_timer() - time0:.1f}s)')

    print('Adding people aliases to graph...')
    time1 = timeit.default_timer()
    last_node, last_edge, nodes, edges = people2graphml(node_cnt=last_idx + 1,
                                                        edge_cnt=1,
                                                        idx2ppl=idx2people,
                                                        graphml_nodes='',
                                                        graphml_edges='')
    print(f' - done after {timeit.default_timer() - time1:.1f}s (total: {timeit.default_timer() - time0:.1f}s)')

    print('Constructing block index to identify duplicates...')
    time1 = timeit.default_timer()
    email_index, email_blocks, people, last_node, last_edge, nodes, edges = \
        build_block_index(quagga_file=args.quagga_parsed,
                          ppl=people,
                          node_cnt=last_node,
                          edge_cnt=last_edge,
                          graphml_nodes=nodes,
                          graphml_edges=edges)
    print(f' - done after {timeit.default_timer() - time1:.1f}s (total: {timeit.default_timer() - time0:.1f}s)')

    print('Adding parsed email blocks to graph...')
    time1 = timeit.default_timer()
    nodes, edges = quagga2graphml(blocks=email_blocks,
                                  block_index=email_index,
                                  node_cnt=last_node,
                                  edge_cnt=last_edge,
                                  graphml_nodes=nodes,
                                  graphml_edges=edges,
                                  include_text=args.include_text)
    print(f' - done after {timeit.default_timer() - time1:.1f}s (total: {timeit.default_timer() - time0:.1f}s)')

    print('Constructing GraphML in memory...')
    time1 = timeit.default_timer()
    final_graphml = construct_graphml(graphml_nodes=nodes, graphml_edges=edges)
    print(f' - done after {timeit.default_timer() - time1:.1f}s (total: {timeit.default_timer() - time0:.1f}s)')

    print('Writing GraphML to disk...')
    time1 = timeit.default_timer()
    args.graph.write(final_graphml)
    print(f' - done after {timeit.default_timer() - time1:.1f}s (total: {timeit.default_timer() - time0:.1f}s)')

    print('All done!')
    print('Done at', time.strftime("%Y-%m-%d %H:%M"))
